import pytest
import subprocess
import os
import json

import laspy
from tests.test_common import test1_4_las, autzen_las

try:
    import pyproj
except ModuleNotFoundError:
    pyproj = None


@pytest.fixture()
def file_wkt():
    return laspy.read(test1_4_las)


@pytest.fixture()
def file_geotiff():
    return laspy.read(autzen_las)

def has_pyproj() -> bool:
    return pyproj is not None


@pytest.mark.skipif(
    not has_pyproj(), reason="pyproj not installed"
)
def test_parse_crs_wkt(file_wkt):
    assert 'Bound CRS' in file_wkt.header.parse_crs().type_name


@pytest.mark.skipif(
    not has_pyproj(), reason="pyproj not installed"
)
def test_parse_crs_geotiff(file_geotiff):
    assert 'Projected CRS' in file_geotiff.header.parse_crs().type_name


@pytest.mark.skipif(
    not has_pyproj(), reason="pyproj not installed"
)
def test_add_crs_wkt():
    header = laspy.LasHeader(point_format=6, version="1.4")
    crs = pyproj.CRS.from_epsg(6432)
    header.add_crs(crs)

    lasf_proj = header.vlrs.get_by_id('LASF_Projection')

    assert len(lasf_proj) > 0
    assert hasattr(lasf_proj[0], 'string')
    assert lasf_proj[0].string == crs.to_wkt()
    assert header.global_encoding.wkt


@pytest.mark.skipif(
    not has_pyproj(), reason="pyproj not installed"
)
def test_add_crs_geotiff():
    """
    Test adding geotiff crs seems ok
    """
    header = laspy.LasHeader(point_format=3, version="1.2")
    crs = pyproj.CRS.from_epsg(6432)
    header.add_crs(crs)

    lasf_proj = header.vlrs.get_by_id('LASF_Projection')

    assert len(lasf_proj) > 0
    assert hasattr(lasf_proj[0], 'geo_keys') 
    assert hasattr(lasf_proj[1], 'strings')
    assert header.global_encoding.wkt == False


    las = laspy.LasData(header=header)
    las = laspy.lib.write_then_read_again(las)
    epsg = las.header.parse_crs().to_epsg()
    assert epsg == 6432, "epsg after read-write is not the same"


@pytest.mark.skipif(
    not has_pyproj(), reason="pyproj not installed"
)
def test_add_crs_compatibility():
    header = laspy.LasHeader(point_format=3, version="1.4")
    crs = pyproj.CRS.from_epsg(6432)
    header.add_crs(crs, keep_compatibility=False)

    lasf_proj = header.vlrs.get_by_id('LASF_Projection')

    assert len(lasf_proj) > 0
    assert hasattr(lasf_proj[0], 'string')
    assert lasf_proj[0].string == crs.to_wkt()
    assert header.global_encoding.wkt

def has_pdal():
    try:
        subprocess.run(['pdal', '--version'], capture_output=True, check=True)
    except:
        return False
    else:
        return True


def get_pdal_srs(filename):
    process = subprocess.run(['pdal', 'info', '--metadata', filename], capture_output=True, check=True)
    json_output = json.loads(process.stdout.decode('utf-8'))['metadata']['srs']
    return json_output['wkt'], json_output['proj4']
    

def geotiff_crs_pdal_test(crs):
    header = laspy.LasHeader(point_format=3, version="1.2")
    header.add_crs(crs)

    tmp_filename = f'delete_me_pdal_test_cs_{crs.to_epsg()}.las'

    las = laspy.LasData(header=header)
    las.write(tmp_filename)


    try:
        pdal_wkt, _ = get_pdal_srs(tmp_filename)

    finally:
        os.remove(tmp_filename)


    # We don't use pdal's proj4 as its seems not complete enough
    # to be 100% equal, while the wkt seems ok
    assert pyproj.CRS.from_wkt(pdal_wkt) == crs


@pytest.mark.skipif(
    not (has_pdal() and has_pyproj()), reason="PDAL and/or pyproj not found"
)
def test_pdal_understands_our_geotiff_derived_projected_cs():
    crs = pyproj.CRS.from_epsg(3945)
    geotiff_crs_pdal_test(crs)
    
@pytest.mark.skipif(
    not (has_pdal() and has_pyproj()), reason="PDAL and/or pyproj not found"
)
def test_pdal_understands_our_geotiff_geographic_cs():
    crs = pyproj.CRS.from_epsg(4326)
    geotiff_crs_pdal_test(crs)



@pytest.mark.skipif(
    not (has_pdal() and has_pyproj()), reason="PDAL and/or pyproj not found"
)
def test_pdal_understands_our_geotiff_projected_cs():
    crs = pyproj.CRS.from_epsg(6432)
    geotiff_crs_pdal_test(crs)




__version__ = "2.2.0b0"

import logging

from . import errors, vlrs, file
from .errors import LaspyException
from .laswriter import LasWriter
from .lasreader import LasReader
from .lib import LazBackend, convert
from .lib import create_las as create
from .lib import mmap_las as mmap
from .lib import open_las as open
from .lib import read_las as read
from .point import PointFormat, ExtraBytesParams, DimensionKind, DimensionInfo
from .point.record import PackedPointRecord, ScaleAwarePointRecord
from .point.dims import supported_point_formats, supported_versions
from .point.format import lost_dimensions
from .header import LasHeader
from .lasdata import LasData
from .vlrs import VLR

logging.getLogger(__name__).addHandler(logging.NullHandler())



from pathlib import Path

import numpy as np
import pytest

import laspy
from laspy.lib import write_then_read_again
from . import conftest

simple_las = conftest.SIMPLE_LAS_FILE_PATH
simple_laz = conftest.SIMPLE_LAZ_FILE_PATH
vegetation1_3_las = conftest.VEGETATION1_3_LAS_FILE_PATH
test1_4_las = conftest.TEST1_4_LAS_FILE_PATH
extra_bytes_las = conftest.EXTRA_BYTES_LAS_FILE_PATH
extra_bytes_laz = conftest.EXTRA_BYTES_LAZ_FILE_PATH
plane_laz = conftest.PLANE_LAZ_FILE_PATH
autzen_las = conftest.AUTZEN_FILE_PATH

if not laspy.LazBackend.detect_available():
    do_compression = [False]
    all_file_paths = [simple_las, vegetation1_3_las, test1_4_las, extra_bytes_las]
else:
    do_compression = [False, True]
    all_file_paths = [
        simple_las,
        simple_laz,
        vegetation1_3_las,
        test1_4_las,
        plane_laz,
        extra_bytes_laz,
        extra_bytes_las,
    ]


@pytest.fixture(params=all_file_paths)
def las(request):
    return laspy.read(request.param)


@pytest.fixture(params=[simple_las, vegetation1_3_las])
def all_las_but_1_4(request):
    return laspy.read(request.param)


@pytest.fixture(params=[simple_las, vegetation1_3_las, test1_4_las, extra_bytes_las])
def las_path_fixture(request):
    return request.param


@pytest.fixture(params=[simple_laz, extra_bytes_laz, plane_laz])
def all_laz_path(request):
    return request.param


def dim_does_not_exists(las, dim_name):
    try:
        _ = getattr(las, dim_name)
    except AttributeError:
        return True
    return False


def dim_does_exists(las, dim_name):
    try:
        _ = getattr(las, dim_name)
    except AttributeError:
        return False
    return True


def test_change_format(las):
    in_version = las.header.version

    las = laspy.convert(las, point_format_id=2)
    las = write_then_read_again(las)
    assert las.points.point_format.id == 2
    assert las.header.point_format.id == 2
    assert las.header.version == in_version
    assert dim_does_not_exists(las, "gps_time")

    las = laspy.convert(las, point_format_id=1)
    las = write_then_read_again(las)
    assert las.points.point_format.id == 1
    assert las.header.point_format.id == 1
    assert las.header.version == in_version
    assert dim_does_not_exists(las, "red")
    assert dim_does_not_exists(las, "green")
    assert dim_does_not_exists(las, "blue")

    las = laspy.convert(las, point_format_id=0)
    las = write_then_read_again(las)
    assert las.points.point_format.id == 0
    assert las.header.point_format.id == 0
    assert las.header.version == in_version
    assert dim_does_not_exists(las, "red")
    assert dim_does_not_exists(las, "green")
    assert dim_does_not_exists(las, "blue")
    assert dim_does_not_exists(las, "gps_time")

    las = laspy.convert(las, point_format_id=8)
    las = write_then_read_again(las)
    assert str(las.header.version) == "1.4"
    assert las.points.point_format.id == 8
    assert las.header.point_format.id == 8
    assert dim_does_exists(las, "red")
    assert dim_does_exists(las, "green")
    assert dim_does_exists(las, "blue")
    assert dim_does_exists(las, "nir")

    las = laspy.convert(las, point_format_id=7)
    las = write_then_read_again(las)
    assert str(las.header.version) == "1.4"
    assert las.points.point_format.id == 7
    assert las.header.point_format.id == 7
    assert dim_does_exists(las, "red")
    assert dim_does_exists(las, "green")
    assert dim_does_exists(las, "blue")
    assert dim_does_not_exists(las, "nir")

    las = laspy.convert(las, point_format_id=6)
    las = write_then_read_again(las)
    assert str(las.header.version) == "1.4"
    assert las.points.point_format.id == 6
    assert las.header.point_format.id == 6
    assert dim_does_not_exists(las, "red")
    assert dim_does_not_exists(las, "green")
    assert dim_does_not_exists(las, "blue")
    assert dim_does_not_exists(las, "nir")


def test_rw_all_set_one(las):
    for dim_name in las.point_format.dimension_names:
        las[dim_name][:] = 1

    for dim_name in las.point_format.dimension_names:
        assert np.alltrue(las[dim_name] == 1), "{} not equal".format(dim_name)

    las2 = write_then_read_again(las)

    for dim_name in las.point_format.dimension_names:
        assert np.alltrue(las[dim_name] == las2[dim_name]), "{} not equal".format(
            dim_name
        )


def test_coords_do_not_break(las):
    xs, ys, zs = las.x, las.y, las.z

    las.x = xs
    las.y = ys
    las.z = zs

    assert np.allclose(xs, las.x)
    assert np.allclose(ys, las.y)
    assert np.allclose(zs, las.z)


def test_coords_when_setting_offsets_and_scales(las):
    new_las = laspy.create()

    new_las.header.offsets = las.header.offsets
    new_las.header.scales = las.header.scales

    new_las.x = las.x
    new_las.y = las.y
    new_las.z = las.z

    assert np.allclose(las.x, new_las.x)
    assert np.allclose(las.y, new_las.y)
    assert np.allclose(las.z, new_las.z)


def test_coords_when_using_create_from_header(las):
    new_las = laspy.LasData(las.header)

    new_las.x = las.x
    new_las.y = las.y
    new_las.z = las.z

    assert np.allclose(las.x, new_las.x)
    assert np.allclose(las.y, new_las.y)
    assert np.allclose(las.z, new_las.z)


def test_slicing(las):
    las.points = las.points[len(las.points) // 2:]


@pytest.mark.parametrize("do_compress", do_compression)
def test_can_write_then_re_read_files(las, do_compress):
    _las = write_then_read_again(las, do_compress=do_compress)


def test_point_record_setitem_scaled_view():
    las = laspy.read(simple_las)
    las.add_extra_dim(
        laspy.ExtraBytesParams(
            "lol", "uint64", scales=np.array([2.0]), offsets=np.array([0.0])
        )
    )

    new_values = np.ones(len(las.points)) * 4
    las.lol = new_values

    assert np.allclose(las.lol, new_values)


def test_laspy_file_raises():
    with pytest.raises(laspy.errors.LaspyException):
        laspy.file.File("some path")


def test_lasdata_setitem_xyz_with_2d_array():
    las = laspy.read(simple_las)

    xyz = np.ones(len(las), dtype='3f8')
    xyz[..., 1] = 2.0
    xyz[..., 2] = 3.0

    las[['x', 'y', 'z']] = xyz

    assert np.all(las.x == xyz[..., 0])
    assert np.all(las.y == xyz[..., 1])
    assert np.all(las.z == xyz[..., 2])


def test_lasdata_setitem_xyz_with_structured_array():
    las = laspy.read(simple_las)

    xyz = np.ones(len(las), dtype=[('x', 'f8'), ('y', 'f8'), ('z', 'f8')])
    xyz['y'] = 2.0
    xyz['z'] = 3.0

    las[['x', 'y', 'z']] = xyz

    assert np.all(las.x == xyz['x'])
    assert np.all(las.y == xyz['y'])
    assert np.all(las.z == xyz['z'])


def test_lasdata_setitem_one_dimension():
    las = laspy.read(simple_las)

    las[['x']] = np.ones(len(las), 'f8') * 17.0
    assert np.all(las.x == 17.0)


def test_lasdata_setitem_works_with_subfields():
    las = laspy.read(simple_las)

    new_values = np.ones(len(las), dtype=[('classification', 'u1'), ('return_number', 'u1')])
    new_values['classification'] = 23
    new_values['return_number'] = 2

    las[['classification', 'return_number']] = new_values
    assert np.all(las.classification == 23)
    assert np.all(las.return_number == 2)


def test_las_data_getitem_indices():
    las = laspy.read(simple_las)
    las.classification[:] = 0

    indices = np.array([1, 2, 3, 4])
    sliced_las = las[indices]
    sliced_las.classification[:] = 1

    assert np.all(sliced_las.classification == 1)
    # We expect sliced_las to give us a copy,
    # So original las is no modified
    assert np.all(las.classification == 0)


def test_las_data_getitem_slice():
    las = laspy.read(simple_las)
    las.classification[:] = 0

    sliced_las = las[0:10]
    sliced_las.classification[:] = 1

    assert np.all(sliced_las.classification == 1)
    # Slicing does not trigger, advanced indexing
    # so the underlying array is not a copy
    # https://numpy.org/doc/stable/reference/arrays.indexing.html
    assert np.all(las.classification[:10] == 1)
    assert np.all(las.classification[10:] == 0)


def test_change_scaling():
    """ Check our change scaling method.

    We expect the scaled x,y,z not to change
    while the unscaled (integers) X,Y,Z should change.
    """
    hdr = laspy.LasHeader()
    hdr.offsets = np.array([0.0, 0.0, 0.0])
    hdr.scales = np.array([1.0, 1.0, 1.0])

    las = laspy.LasData(hdr)

    las['x'] = np.array([1, 2, 3, 4])
    las['y'] = np.array([1, 2, 3, 4])
    las['z'] = np.array([1, 2, 3, 4])

    assert np.all(las.x == [1, 2, 3, 4])
    assert np.all(las.y == [1, 2, 3, 4])
    assert np.all(las.z == [1, 2, 3, 4])

    assert np.all(las.X == [1, 2, 3, 4])
    assert np.all(las.Y == [1, 2, 3, 4])
    assert np.all(las.Z == [1, 2, 3, 4])

    saved_offsets = las.header.offsets.copy()
    las.change_scaling(scales=[0.5, 0.1, 0.01])

    assert np.all(las.header.scales == [0.5, 0.1, 0.01])
    assert np.all(las.header.offsets == saved_offsets)

    assert np.all(las.x == [1, 2, 3, 4])
    assert np.all(las.y == [1, 2, 3, 4])
    assert np.all(las.z == [1, 2, 3, 4])

    assert np.all(las.X == [2, 4, 6, 8])
    assert np.all(las.Y == [10, 20, 30, 40])
    assert np.all(las.Z == [100, 200, 300, 400])

    saved_scales = las.header.scales.copy()
    las.change_scaling(offsets=[1, 20, 30])

    assert np.all(las.header.scales == saved_scales)
    assert np.all(las.header.offsets == [1, 20, 30])

    assert np.all(las.x == [1, 2, 3, 4])
    assert np.all(las.y == [1, 2, 3, 4])
    assert np.all(las.z == [1, 2, 3, 4])

    assert np.all(las.X == [0, 2, 4, 6])
    assert np.all(las.Y == [-190, -180, -170, -160])
    assert np.all(las.Z == [-2900, -2800, -2700, -2600])


def test_setting_x_y_z_on_las_data():
    """
    The goal of this test if to make sure that when setting the `x`,`y` and `z`
    attribute of a LasData object, the X,Y,Z version of the coordinates
    are properly set in the inner point record
    """
    las = laspy.read(simple_las)

    new_las = laspy.create()

    new_las.x = las.x
    new_las.y = las.y
    new_las.z = las.z

    assert np.all(new_las.x == las.x)
    assert np.all(new_las.X == las.X)
    assert np.all(new_las.y == las.y)
    assert np.all(new_las.Y == las.Y)
    assert np.all(new_las.z == las.z)
    assert np.all(new_las.Z == las.Z)


    new_las = laspy.lib.write_then_read_again(new_las)

    assert np.all(new_las.x == las.x)
    assert np.all(new_las.X == las.X)
    assert np.all(new_las.y == las.y)
    assert np.all(new_las.Y == las.Y)
    assert np.all(new_las.z == las.z)
    assert np.all(new_las.Z == las.Z)


def test_setting_xyz_on_las_data():
    """
    The goal of this test if to make sure that when setting the `xyz`
    attribute of a LasData object, the X,Y,Z version of the coordinates
    are properly set in the inner point record
    """
    las = laspy.read(simple_las)

    new_las = laspy.create()

    new_las.xyz = las.xyz

    assert np.all(new_las.x == las.x)
    assert np.all(new_las.X == las.X)
    assert np.all(new_las.y == las.y)
    assert np.all(new_las.Y == las.Y)
    assert np.all(new_las.z == las.z)
    assert np.all(new_las.Z == las.Z)

    new_las = laspy.lib.write_then_read_again(new_las)

    assert np.all(new_las.x == las.x)
    assert np.all(new_las.X == las.X)
    assert np.all(new_las.y == las.y)
    assert np.all(new_las.Y == las.Y)
    assert np.all(new_las.z == las.z)
    assert np.all(new_las.Z == las.Z)



import io
import math
from typing import Union, Iterable, BinaryIO, Optional, cast

import numpy as np

from .compression import LazBackend
from .errors import LaspyException
from .header import LasHeader
from .laswriter import UncompressedPointWriter
from .point.record import PackedPointRecord
from .vlrs.known import LasZipVlr
from .vlrs.vlrlist import VLRList

try:
    import lazrs
except ModuleNotFoundError:
    pass


class LasAppender:
    """Allows to append points to and existing LAS/LAZ file.

    Appending to LAZ is only supported by the lazrs backend
    """

    def __init__(
        self,
        dest: BinaryIO,
        laz_backend: Optional[Union[LazBackend, Iterable[LazBackend]]] = None,
        closefd: bool = True,
        encoding_errors: str = "strict",
    ) -> None:
        if not dest.seekable():
            raise TypeError("Expected the 'dest' to be a seekable file object")
        header = LasHeader.read_from(dest)
        if laz_backend is None:
            laz_backend = [
                bck
                for bck in LazBackend.detect_available()
                if bck is not LazBackend.Laszip
            ]

        self.dest = dest
        self.header = header

        if not header.are_points_compressed:
            self.points_writer = UncompressedPointWriter(self.dest)
            self.dest.seek(
                (self.header.point_count * self.header.point_format.size)
                + self.header.offset_to_point_data,
                io.SEEK_SET,
            )
        else:
            self.points_writer = self._create_laz_backend(laz_backend)

        if header.version.minor >= 4 and header.number_of_evlrs > 0:
            assert (
                self.dest.tell() <= self.header.start_of_first_evlr
            ), "The position is past the start of evlrs"
            pos = self.dest.tell()
            self.dest.seek(self.header.start_of_first_evlr, io.SEEK_SET)
            self.evlrs: Optional[VLRList] = VLRList.read_from(
                self.dest, self.header.number_of_evlrs, extended=True
            )
            dest.seek(self.header.start_of_first_evlr, io.SEEK_SET)
            self.dest.seek(pos, io.SEEK_SET)
        else:
            self.evlrs: Optional[VLRList] = None

        self.closefd = closefd
        self.encoding_errors = encoding_errors

    def append_points(self, points: PackedPointRecord) -> None:
        """Append the points to the file, the points
        must have the same point format as the points
        already contained within the file.

        :param points: The points to append
        """
        if points.point_format != self.header.point_format:
            raise LaspyException("Point formats do not match")

        self.points_writer.write_points(points)
        self.header.grow(points)

    def close(self) -> None:
        self.points_writer.done()
        self._write_evlrs()
        self._write_updated_header()

        if self.closefd:
            self.dest.close()

    def _write_evlrs(self) -> None:
        if (
            self.header.version.minor >= 4
            and self.evlrs is not None
            and len(self.evlrs) > 0
        ):
            self.header.number_of_evlr = len(self.evlrs)
            self.header.start_of_first_evlr = self.dest.tell()
            self.evlrs.write_to(self.dest, as_extended=True)

    def _write_updated_header(self) -> None:
        pos = self.dest.tell()
        self.dest.seek(0, io.SEEK_SET)
        self.header.write_to(
            self.dest, ensure_same_size=True, encoding_errors=self.encoding_errors
        )
        self.dest.seek(pos, io.SEEK_SET)

    def _create_laz_backend(
        self,
        laz_backend: Union[LazBackend, Iterable[LazBackend]] = (
            LazBackend.LazrsParallel,
            LazBackend.Lazrs,
        ),
    ) -> "LazrsAppender":
        try:
            laz_backend = iter(laz_backend)
        except TypeError:
            laz_backend = (laz_backend,)

        last_error = None
        for backend in laz_backend:
            if backend == LazBackend.Laszip:
                raise LaspyException("Laszip backend does not support appending")
            elif backend == LazBackend.LazrsParallel:
                try:
                    return LazrsAppender(self.dest, self.header, parallel=True)
                except Exception as e:
                    last_error = e
            elif backend == LazBackend.Lazrs:
                try:
                    return LazrsAppender(self.dest, self.header, parallel=False)
                except Exception as e:
                    last_error = e

        if last_error is not None:
            raise LaspyException(f"Could not initialize a laz backend: {last_error}")
        else:
            raise LaspyException(f"No valid laz backend selected")

    def __enter__(self) -> "LasAppender":
        return self

    def __exit__(self, exc_type, exc_val, exc_tb) -> None:
        self.close()


class LazrsAppender:
    """Appending in LAZ file
    works by seeking to start of the last chunk
    of compressed points, decompress it while keeping the points in
    memory.

    Then seek back to the start of the last chunk, and recompress
    the points we just read, so that we have a compressor in the proper state
    ready to compress new points.
    """

    def __init__(self, dest: BinaryIO, header: LasHeader, parallel: bool) -> None:
        self.dest = dest
        self.offset_to_point_data = header.offset_to_point_data
        laszip_vlr = cast(LasZipVlr, header.vlrs.get("LasZipVlr")[0])

        self.dest.seek(header.offset_to_point_data, io.SEEK_SET)
        decompressor = lazrs.LasZipDecompressor(self.dest, laszip_vlr.record_data)
        vlr = decompressor.vlr()
        number_of_complete_chunk = int(
            math.floor(header.point_count / vlr.chunk_size())
        )

        if vlr.uses_variable_size_chunks():
            # TODO: this is probably implementable but until its really needed,
            #       or i'm bored, there is no point.
            raise LaspyException(
                "LazrsAppender does not support LAZ files with variable size chunks"
            )

        self.dest.seek(header.offset_to_point_data, io.SEEK_SET)
        chunk_table = [
            byte_count
            for (point_count, byte_count) in lazrs.read_chunk_table(self.dest, vlr)
        ]
        if chunk_table is None:
            # The file does not have a chunk table
            # we cannot seek to the last chunk, so instead, we will
            # decompress points (which is slower) and build the chunk table
            # to write it later

            self.chunk_table = []
            start_of_chunk = self.dest.tell()
            point_buf = bytearray(vlr.chunk_size() * vlr.item_size())

            for _ in range(number_of_complete_chunk):
                decompressor.decompress_many(point_buf)
                pos = self.dest.tell()
                self.chunk_table.append(pos - start_of_chunk)
                start_of_chunk = pos
        else:
            self.chunk_table = chunk_table[:-1]
            idx_first_point_of_last_chunk = number_of_complete_chunk * vlr.chunk_size()
            decompressor.seek(idx_first_point_of_last_chunk)

        num_points_in_last_chunk = header.point_count - (
            number_of_complete_chunk * vlr.chunk_size()
        )
        points_of_last_chunk = bytearray(num_points_in_last_chunk * vlr.item_size())
        decompressor.decompress_many(points_of_last_chunk)

        self.dest.seek(header.offset_to_point_data, io.SEEK_SET)
        if parallel:
            self.compressor = lazrs.ParLasZipCompressor(self.dest, vlr)
        else:
            self.compressor = lazrs.LasZipCompressor(self.dest, vlr)
        # This effectively overwrites the old offset value.
        # But, more importantly, it makes the compressor aware of where to
        # write the offset value when its done.
        self.compressor.reserve_offset_to_chunk_table()
        assert self.dest.tell() == header.offset_to_point_data + 8
        self.dest.seek(sum(self.chunk_table), io.SEEK_CUR)
        self.compressor.compress_many(points_of_last_chunk)
        self.vlr = vlr

    def write_points(self, points: PackedPointRecord) -> None:
        points_bytes = np.frombuffer(points.array, np.uint8)
        self.compressor.compress_many(points_bytes)

    def done(self) -> None:
        # The chunk table written is at the good position
        # but it is incomplete (it's missing the chunk_table of
        # chunks before the one we appended)
        self.compressor.done()

        # Read the chunk table corresponding to our appended chunks
        self.dest.seek(self.offset_to_point_data, io.SEEK_SET)
        appended_chunk_table = [
            byte_count
            for (point_count, byte_count) in lazrs.read_chunk_table(self.dest, self.vlr)
        ]
        chunk_table = self.chunk_table + appended_chunk_table

        # Rewrite the fully complete chunk table.
        self.dest.seek(self.offset_to_point_data, io.SEEK_SET)
        offset_to_chunk_table = int.from_bytes(self.dest.read(8), "little", signed=True)
        self.dest.seek(offset_to_chunk_table, io.SEEK_SET)
        lazrs.write_chunk_table(
            self.dest,
            [(self.vlr.chunk_size(), byte_count) for byte_count in chunk_table],
            self.vlr,
        )



import numpy as np
import pytest

import laspy
from laspy.lib import write_then_read_again


@pytest.mark.parametrize("target_point_format_id", [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10])
def test_point_format_conversion_copies_field_values(file_path, target_point_format_id):
    original = laspy.read(file_path)
    converted = laspy.convert(original, point_format_id=target_point_format_id)
    converted = write_then_read_again(converted)

    converted_dimension_names = set(converted.point_format.dimension_names)
    dimension_expected_to_be_kept = [
        dim_name
        for dim_name in original.point_format.dimension_names
        if dim_name in converted_dimension_names
    ]

    for dim_name in dimension_expected_to_be_kept:
        assert np.allclose(
            converted[dim_name], original[dim_name]
        ), "{} not equal".format(dim_name)



import abc
import io
import logging
from typing import Optional, BinaryIO, Iterable, Union

from . import errors
from .compression import LazBackend
from .header import LasHeader
from .lasdata import LasData
from .point import record
from .vlrs.known import LasZipVlr
from .vlrs.vlrlist import VLRList

try:
    import lazrs
except ModuleNotFoundError:
    pass

try:
    import laszip
except ModuleNotFoundError:
    pass

logger = logging.getLogger(__name__)


class LasReader:
    """The reader class handles LAS and LAZ via one of the supported backend"""

    def __init__(
        self,
        source: BinaryIO,
        closefd: bool = True,
        laz_backend: Optional[Union[LazBackend, Iterable[LazBackend]]] = None,
    ):
        """
        Initialize the LasReader

        Parameters
        ----------
        source: file_object
        closefd: bool, default True
        laz_backend: LazBackend or list of LazBackend, optional
        """
        self.closefd = closefd
        if laz_backend is None:
            laz_backend = LazBackend.detect_available()
        self.laz_backend = laz_backend
        self.header = LasHeader.read_from(source)

        if self.header.point_count > 0:
            if self.header.are_points_compressed:
                self.point_source = self._create_laz_backend(source)
                if self.point_source is None:
                    raise errors.LaspyException(
                        "Data is compressed, but no LazBacked could be initialized"
                    )
            else:
                self.point_source = UncompressedPointReader(source, self.header)
        else:
            self.point_source = EmptyPointReader()

        self.points_read = 0

    def read_points(self, n: int) -> record.ScaleAwarePointRecord:
        """Read n points from the file


        Will only read as many points as the header advertise.
        That is, if you ask to read 50 points and there are only 45 points left
        this function will only read 45 points.

        If there are no points left to read, returns an empty point record.

        Parameters
        ----------
        n: The number of points to read
           if n is less than 0, this function will read the remaining points
        """
        points_left = self.header.point_count - self.points_read
        if points_left <= 0:
            return record.ScaleAwarePointRecord.empty(
                self.header.point_format,
                self.header.scales,
                self.header.offsets,
            )

        if n < 0:
            n = points_left
        else:
            n = min(n, points_left)

        r = record.PackedPointRecord.from_buffer(
            self.point_source.read_n_points(n), self.header.point_format
        )
        if len(r) < n:
            logger.error(f"Could only read {len(r)} of the requested {n} points")

        points = record.ScaleAwarePointRecord(
            r.array, r.point_format, self.header.scales, self.header.offsets
        )

        self.points_read += n
        return points

    def read(self) -> LasData:
        """Reads all the points not read and returns a LasData object"""
        points = self.read_points(-1)
        las_data = LasData(header=self.header, points=points)

        if self.header.version.minor >= 4:
            if (
                self.header.are_points_compressed
                and not self.point_source.source.seekable()
            ):
                # We explicitly require seekable stream because we have to seek
                # past the chunk table of LAZ file
                raise errors.LaspyException(
                    "source must be seekable, to read evlrs form LAZ file"
                )
            self.point_source.source.seek(self.header.start_of_first_evlr, io.SEEK_SET)
            las_data.evlrs = self._read_evlrs(self.point_source.source, seekable=True)

        return las_data

    def seek(self, pos: int, whence: int = io.SEEK_SET) -> int:
        """Seeks to the start of the point at the given pos

        Parameters
        ----------
        pos: index of the point to seek to
        whence: optional, controls how the pos parameter is interpreted:
                io.SEEK_SET: (default) pos is the index of the point from the beginning
                io.SEEK_CUR: pos is the point_index relative to the point_index of the last point read
                io.SEEK_END: pos is the point_index relative to last point
        Returns
        -------
        The index of the point the reader seeked to, relative to the first point
        """
        if whence == io.SEEK_SET:
            allowed_range = range(0, self.header.point_count)
            point_index = pos
        elif whence == io.SEEK_CUR:
            allowed_range = range(
                -self.points_read, self.header.point_count - self.points_read
            )
            point_index = self.points_read + pos
        elif whence == io.SEEK_END:
            allowed_range = range(-self.header.point_count, 0)
            point_index = self.header.point_count + pos
        else:
            raise ValueError(f"Invalid value for whence: {whence}")

        if pos not in allowed_range:
            whence_str = ["start", "current point", "end"]
            raise IndexError(
                f"When seeking from the {whence_str[whence]}, pos must be in {allowed_range}"
            )

        self.point_source.seek(point_index)
        self.points_read = point_index
        return point_index

    def chunk_iterator(self, points_per_iteration: int) -> "PointChunkIterator":
        """Returns an iterator, that will read points by chunks
        of the requested size

        :param points_per_iteration: number of points to be read with each iteration
        :return:
        """
        return PointChunkIterator(self, points_per_iteration)

    def close(self) -> None:
        """closes the file object used by the reader"""
        if self.closefd:
            self.point_source.close()

    def _create_laz_backend(self, source) -> Optional["IPointReader"]:
        """Creates the laz backend to use according to `self.laz_backend`.

        If `self.laz_backend` contains mutilple backends, this functions will
        try to create them in order until one of them is successfully constructed.

        If none could be constructed, the error of the last backend tried wil be raised
        """
        if not self.laz_backend:
            raise errors.LaspyException(
                "No LazBackend selected, cannot decompress data"
            )

        try:
            backends = iter(self.laz_backend)
        except TypeError:
            backends = (self.laz_backend,)

        laszip_vlr = self.header.vlrs.pop(self.header.vlrs.index("LasZipVlr"))
        last_error = None
        for backend in backends:
            try:
                if not backend.is_available():
                    raise errors.LaspyException(f"The '{backend}' is not available")

                if backend == LazBackend.LazrsParallel:
                    return LazrsPointReader(source, laszip_vlr, parallel=True)
                elif backend == LazBackend.Lazrs:
                    return LazrsPointReader(source, laszip_vlr, parallel=False)
                elif backend == LazBackend.Laszip:
                    return LaszipPointReader(source, self.header)
                else:
                    raise errors.LaspyException(
                        "Unknown LazBackend: {}".format(backend)
                    )

            except Exception as e:
                last_error = e
                logger.error(e)

        raise last_error

    def _read_evlrs(self, source, seekable=False) -> Optional[VLRList]:
        """Reads the EVLRs of the file, will fail if the file version
        does not support evlrs
        """
        if (
            self.header.version.minor >= 4
            and self.points_read == self.header.point_count
        ):
            if seekable:
                source.seek(self.header.start_of_first_evlr)
            return VLRList.read_from(source, self.header.number_of_evlrs, extended=True)
        else:
            return None

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()


class PointChunkIterator:
    def __init__(self, reader: LasReader, points_per_iteration: int) -> None:
        self.reader = reader
        self.points_per_iteration = points_per_iteration

    def __next__(self) -> record.ScaleAwarePointRecord:
        points = self.reader.read_points(self.points_per_iteration)
        if not points:
            raise StopIteration
        return points

    def __iter__(self) -> "PointChunkIterator":
        return self


class IPointReader(abc.ABC):
    """The interface to be implemented by the class that actually reads
    points from as LAS/LAZ file so that the LasReader can use it.

    It is used to manipulate LAS/LAZ (with different LAZ backends) in the
    reader
    """

    @abc.abstractmethod
    def read_n_points(self, n: int) -> bytearray:
        ...

    @abc.abstractmethod
    def seek(self, point_index: int) -> None:
        ...

    @abc.abstractmethod
    def close(self) -> None:
        ...


class UncompressedPointReader(IPointReader):
    """Implementation of IPointReader for the simple uncompressed case"""

    def __init__(self, source, header: LasHeader) -> None:
        self.source = source
        self.header = header

    def read_n_points(self, n: int) -> bytearray:
        try:
            readinto = self.source.readinto
        except AttributeError:
            data = bytearray(self.source.read(n * self.header.point_format.size))
        else:
            data = bytearray(n * self.header.point_format.size)
            num_read = readinto(data)
            if num_read < len(data):
                data = data[:num_read]

        return data

    def seek(self, point_index: int) -> None:
        self.source.seek(
            self.header.offset_to_point_data
            + (point_index * self.header.point_format.size)
        )

    def close(self):
        self.source.close()


class LaszipPointReader(IPointReader):
    """Implementation for the laszip backend"""

    def __init__(self, source: BinaryIO, header: LasHeader) -> None:
        self.source = source
        self.source.seek(0)
        self.unzipper = laszip.LasUnZipper(source)
        unzipper_header = self.unzipper.header
        assert unzipper_header.point_data_format == header.point_format.id
        assert unzipper_header.point_data_record_length == header.point_format.size
        self.point_size = header.point_format.size

    def read_n_points(self, n: int) -> bytearray:
        points_data = bytearray(n * self.point_size)
        self.unzipper.decompress_into(points_data)
        return points_data

    def seek(self, point_index: int) -> None:
        self.unzipper.seek(point_index)

    def close(self) -> None:
        self.source.close()


class LazrsPointReader(IPointReader):
    """Implementation for the laz-rs backend, supports single-threaded decompression
    as well as multi-threaded decompression
    """

    def __init__(self, source, laszip_vlr: LasZipVlr, parallel: bool) -> None:
        self.source = source
        self.vlr = lazrs.LazVlr(laszip_vlr.record_data)
        if parallel:
            self.decompressor = lazrs.ParLasZipDecompressor(
                source, laszip_vlr.record_data
            )
        else:
            self.decompressor = lazrs.LasZipDecompressor(source, laszip_vlr.record_data)

    def read_n_points(self, n: int) -> bytearray:
        point_bytes = bytearray(n * self.vlr.item_size())
        self.decompressor.decompress_many(point_bytes)
        return point_bytes

    def seek(self, point_index: int) -> None:
        self.decompressor.seek(point_index)

    def close(self) -> None:
        self.source.close()


class EmptyPointReader(IPointReader):
    """Does nothing but returning empty bytes.
    Used to make sure we handle empty LAS files in a robust way.
    """

    def read_n_points(self, n: int) -> bytearray:
        return bytearray()

    def close(self) -> None:
        pass

    def seek(self, point_index: int) -> None:
        pass
